import { injectable, inject } from 'inversify';
import { IConnection, Location } from 'vscode-languageserver';
import {
    Diagnostic, CompletionList, Hover,
    SymbolInformation, DocumentSymbolParams, DocumentHighlight,
} from 'vscode-languageserver-types';
import { getCSSLanguageService, LanguageService, Stylesheet } from 'vscode-css-languageservice';
import { TextDocumentPositionParams } from 'vscode-base-languageclient';
import { TextDocument, uriToPath, convertChangesRange } from '@workbench-stack/language-core';
import * as lodash from 'lodash';
import { logger } from '@cdm-logger/server';

@injectable()
export class CssService {
    private activated: () => void;
    private activation = new Promise<void>((resolve, reject) => {
        this.activated = resolve;
    });
    private openedDocumentUris: Map<string, any> = new Map<string, any>();
    protected readonly cssService: LanguageService = getCSSLanguageService();

    protected readonly pendingValidationRequests = new Map<string, number>();
    public readonly documents;

    constructor() {
        this.activated();
    }

    public onInitialize = () => {
        logger.debug('The css server is initialized.');

        return {
            capabilities: {
                completionProvider: true,
                hoverProvider: false,
                documentSymbolProvider: true,
                referencesProvider: true,
                definitionProvider: true,
                documentHighlightProvider: true,
                renameProvider: true,
                colorProvider: true,
            },
        };
    }

    public onDidOpenTextDocument = async params => {
        logger.debug('onDidOpenTextDocument css');
        await this.activation;
        const path = uriToPath(params.textDocument.uri);
        const document = new TextDocument(params.textDocument);
        this.openedDocumentUris.set(params.textDocument.uri, document);
    }

    public onDidChangeTextDocument = async params => {
        logger.debug('onDidChangeTextDocument css');
        await this.activation;
        const document = this.getOpenDocument(params.textDocument.uri);
        const changes = convertChangesRange(params.contentChanges);
        document.apply(changes, params.textDocument.version);
    }

    public onDidCloseTextDocument = async (params) => {
        await this.activation;
        const path = uriToPath(params.textDocument.uri);
        this.openedDocumentUris.delete(params.textDocument.uri);
    }

    public onReferences(params): Location[] {
        let document = this.getOpenDocument(params.textDocument.uri) as any;
        const cssDocument = this.getCSSDocument(document);
        return this.cssService.findReferences(document, params.position, cssDocument);
    }

    public onDocumentHighlight(params): DocumentHighlight[] {
        let document = this.getOpenDocument(params.textDocument.uri) as any;
        const cssDocument = this.getCSSDocument(document);
        return this.cssService.findDocumentHighlights(document, params.position, cssDocument);
    }

    public onDefinition(params): Location {
        logger.debug('onDefinition css');
        let document = this.getOpenDocument(params.textDocument.uri) as any;
        const cssDocument = this.getCSSDocument(document);

        const res = this.cssService.findDefinition(document, params.position, cssDocument);
        if (res === null) {
            return {
                uri: params.textDocument.uri,
                range: null,
            };
        }
        return res;
    }
    public onDocumentSymbol(params: DocumentSymbolParams): SymbolInformation[] {
        const document = this.getOpenDocument(params.textDocument.uri) as any;
        const cssDocument = this.getCSSDocument(document);
        const res = this.cssService.findDocumentSymbols(document, cssDocument);

        return res.map(item => {
            if (!item.containerName) {
                item.containerName = '';
            }
            return item;
        });
    }

    public onHover(params: TextDocumentPositionParams): Hover | null {
        const document = this.getOpenDocument(params.textDocument.uri) as any;
        const cssDocument = this.getCSSDocument(document);
        const res: any = this.cssService.doHover(document, params.position, cssDocument);
        if (res) {
            res.contents = lodash.map(res.contents, item => {
                return {
                    value: item,
                    language: 'json',
                };
            }) as any;
        }
        return res;
    }

    public onCompletion(params: TextDocumentPositionParams): CompletionList | null {
        const document = this.getOpenDocument(params.textDocument.uri) as any;
        const cssDocument = this.getCSSDocument(document);
        const res = this.cssService.doComplete(document, params.position, cssDocument);
        res.items = res.items.map(item => {
            if (!item.detail) {
                item.detail = '';
            }
            return item;
        });
        return res;
    }

    public getCSSDocument(document): Stylesheet {
        return this.cssService.parseStylesheet(document);
    }

    private getOpenDocument(uri: string): TextDocument {
        const doc = this.openedDocumentUris.get(uri);
        if (doc) {
            return doc;
        } else {
            throw new Error('Document ' + uri + ' has not been opened.');
        }
    }

}
