import { injectable, inject } from 'inversify';
import * as fs from 'fs';
import { xhr, XHRResponse, getErrorStatusDescription } from 'request-light';
import Uri from 'vscode-uri';
import {
    Diagnostic, Command, CompletionList, CompletionItem, Hover,
    SymbolInformation, DocumentSymbolParams, TextEdit,
} from 'vscode-languageserver-types';
import { getLanguageService, LanguageService, JSONDocument } from 'vscode-json-languageservice';
import { TextDocumentPositionParams, DocumentRangeFormattingParams, ExecuteCommandParams, CodeActionParams } from 'vscode-base-languageclient';
import { TextDocuments } from 'vscode-languageserver';
import { TextDocument, uriToPath, convertChangesRange } from '@workbench-stack/language-core';
import * as lodash from 'lodash';
import { logger } from '@cdm-logger/server';

@injectable()
export class JsonService {
    private openedDocumentUris: Map<string, any> = new Map<string, any>();

    private activated: () => void;
    private activation = new Promise<void>((resolve, reject) => {
        this.activated = resolve;
    });

    protected readonly jsonService: LanguageService = getLanguageService({
        schemaRequestService: this.resovleSchema.bind(this),
    });

    protected readonly pendingValidationRequests = new Map<string, number>();
    public readonly documents = new TextDocuments();
    constructor() {
        this.activated();
    }

    public onInitialize(params) {
        return {
            capabilities: {
                completionProvider: true,
                hoverProvider: true,
                documentSymbolProvider: true,
                documentFormattingProvider: true,

                textDocumentSync: false,
                definitionProvider: false,
                referencesProvider: false,
                workspaceSymbolProvider: false,
                documentHighlightProvider: false,
                documentLinkProvider: false,
                renameProvider: false,
            },
        };
    }

    public onDidOpenTextDocument = async params => {
        logger.debug('onDidOpenTextDocument json');
        await this.activation;
        const path = uriToPath(params.textDocument.uri);
        const document = new TextDocument(params.textDocument);
        this.openedDocumentUris.set(params.textDocument.uri, document);
    }

    public onDidChangeTextDocument = async params => {
        logger.debug('onDidChangeTextDocument json');
        await this.activation;
        const document = this.getOpenDocument(params.textDocument.uri);
        const changes = convertChangesRange(params.contentChanges);
        document.apply(changes, params.textDocument.version);
    }

    public onDidCloseTextDocument = async (params) => {
        await this.activation;
        const path = uriToPath(params.textDocument.uri);
        this.openedDocumentUris.delete(params.textDocument.uri);
    }

    public onDocumentFormatting(params: DocumentRangeFormattingParams): TextEdit[] {
        logger.debug('format json');
        const document = this.getOpenDocument(params.textDocument.uri) as any;
        return this.jsonService.format(document, params.range, params.options);
    }

    public onDocumentSymbol(params: DocumentSymbolParams): any[] {
        logger.debug('findDocumentSymbols json');
        const document = this.getOpenDocument(params.textDocument.uri) as any;
        const jsonDocument = this.getJSONDocument(document);

        const jsonDocumentEditor = lodash.clone(document);
        jsonDocumentEditor.uri = document.uri._path;
        const res = this.jsonService.findDocumentSymbols(jsonDocumentEditor, jsonDocument);

        return res.map(item => {
            if (!item.containerName) {
                item.containerName = '';
            }
            return item;
        });
    }

    public executeCommand(params: ExecuteCommandParams, executeCommandExported): any {
        executeCommandExported(params, this.documents);
    }

    public async onHover(params: TextDocumentPositionParams): Promise<any> {
        logger.debug('onHover');
        const document = this.getOpenDocument(params.textDocument.uri) as any;
        const jsonDocument = this.getJSONDocument(document);
        const {line, character} = params.position;
        const res: any = await this.jsonService.doHover(document, params.position, jsonDocument);

        if (!res) { return {contents: [], range: { start: {}, end: {}}}; }
        res.contents = lodash.map(res.contents, item => {
            return {
                value: item,
                language: 'json',
            };
        }) as any;
        return res;
    }

    public resovleSchema(url: string): Promise<string> {
        logger.debug('resovleSchema  json');
        const uri = Uri.parse(url);
        if (uri.scheme === 'file') {
            return new Promise<string>((resolve, reject) => {
                fs.readFile(uri.fsPath, 'UTF-8', (err, result) => {
                    err ? reject('') : resolve(result.toString());
                });
            });
        }
        return xhr({ url, followRedirects: 5 }).then(response => {
            return response.responseText;
        }, (error: XHRResponse) => {
            return Promise.reject(error.responseText || getErrorStatusDescription(error.status) || error.toString());
        });
    }

    public resolveCompletion(item: CompletionItem): Thenable<any> {
        logger.debug('resolveCompletion  json');
        return this.jsonService.doResolve(item);
    }

    public onCompletion(params: TextDocumentPositionParams): Thenable<any> {
        logger.debug('completion  json');
        const document = this.getOpenDocument(params.textDocument.uri) as any;
        const jsonDocument = this.getJSONDocument(document);
        return this.jsonService.doComplete(document, params.position, jsonDocument);
    }

    public getJSONDocument(document): JSONDocument {
        return this.jsonService.parseJSONDocument(document);
    }

    private getOpenDocument(uri: string): TextDocument {
        const doc = this.openedDocumentUris.get(uri);
        if (doc) {
            return doc;
        } else {
            throw new Error('Document ' + uri + ' has not been opened.');
        }
    }

    public onDefinition() {
        return {
            uri: 'some',
            range: {
                line: 0,
                character: 0,
            },
        };
    }

}
